/**
 * \file	des.h
 * @brief	high-speed implementation of DES
 *
 * - Copyright : Samsung Electronics CO.LTD.,
 *
 * \internal
 * Author : Jisoon Park
 * Dept : DRM Lab, Digital Media Laboratory
 * Creation date : 2006/12/06
 */

#ifndef _DES_H
#define _DES_H

////////////////////////////////////////////////////////////////////////////
// Include Header Files
////////////////////////////////////////////////////////////////////////////
#include "CryptoCore.h"

////////////////////////////////////////////////////////////////////////////
// Define Constants
////////////////////////////////////////////////////////////////////////////
#define SDRM_DES_NUM_OF_ROUNDS	16

////////////////////////////////////////////////////////////////////////////
// Macros
////////////////////////////////////////////////////////////////////////////
/*!	@brief	circular left and right shifts for cc_u32	*/
#ifdef _MSC_VER
	#define SDRM_rotl32(A, B)	_lrotl((A), (B))
	#define SDRM_rotr32(A, B)	_lrotr((A), (B))
#else
	#define SDRM_rotl32(A, B) (((A) << (B)) + ((A) >> (32 - (B))))
	#define SDRM_rotr32(A, B) (((A) >> (B)) + ((A) << (32 - (B))))
#endif

/*!	@brief	permutation operation						*/
#define SDRM_PERM_OP(a,b,t,n,m) {				\
	(t) = ((((a)>>(n))^(b))&(m));				\
	(b) ^= (t);									\
	(a) ^= ((t)<<(n));							\
}

/*!	@brief	initial permutation							*/
#define SDRM_IP(l,r) {							\
	cc_u32 tt;									\
	SDRM_PERM_OP(r,l,tt, 4,0x0f0f0f0f);			\
	SDRM_PERM_OP(l,r,tt,16,0x0000ffff);			\
	SDRM_PERM_OP(r,l,tt, 2,0x33333333);			\
	SDRM_PERM_OP(l,r,tt, 8,0x00ff00ff);			\
	SDRM_PERM_OP(r,l,tt, 1,0x55555555);			\
}

/*!	@brief	inverse of initial permutation				*/
#define SDRM_INV_IP(l,r) {						\
	cc_u32 tt;									\
	SDRM_PERM_OP(l,r,tt, 1,0x55555555);			\
	SDRM_PERM_OP(r,l,tt, 8,0x00ff00ff);			\
	SDRM_PERM_OP(l,r,tt, 2,0x33333333);			\
	SDRM_PERM_OP(r,l,tt,16,0x0000ffff);			\
	SDRM_PERM_OP(l,r,tt, 4,0x0f0f0f0f);			\
}

/*!	@brief	encrypt one round							*/
#define SDRM_D_ENCRYPT(L,R) {					\
	u = R ^ RoundKey[i][0];						\
	t = R ^ RoundKey[i][1];						\
	t = SDRM_rotr32(t, 4);						\
	L^=	SDRM_DES_SPtrans[0][(u >>  2U) & 0x3f]^	\
		SDRM_DES_SPtrans[2][(u >> 10U) & 0x3f]^	\
		SDRM_DES_SPtrans[4][(u >> 18U) & 0x3f]^	\
		SDRM_DES_SPtrans[6][(u >> 26U) & 0x3f]^	\
		SDRM_DES_SPtrans[1][(t >>  2U) & 0x3f]^	\
		SDRM_DES_SPtrans[3][(t >> 10U) & 0x3f]^	\
		SDRM_DES_SPtrans[5][(t >> 18U) & 0x3f]^	\
		SDRM_DES_SPtrans[7][(t >> 26U) & 0x3f];	\
}

////////////////////////////////////////////////////////////////////////////
// static values - specified in FIPS 46
////////////////////////////////////////////////////////////////////////////
static const cc_u8 SDRM_DES_KS_SHIFT[16] = {1, 1, 2, 2, 2, 2, 2, 2, 1, 2, 2, 2, 2, 2, 2, 1};

static const cc_u32 SDRM_DES_BitMask[] = {
	   0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01
};

static const cc_u8 SDRM_DES_KS_PC1[] = {
	56, 48, 40, 32, 24, 16,  8,
	 0, 57, 49, 41, 33, 25, 17,
	 9,  1, 58, 50, 42, 34, 26,
	18, 10,  2, 59, 51, 43, 35,

	62, 54, 46, 38, 30, 22, 14,
	 6, 61, 53, 45, 37, 29, 21,
	13,  5, 60, 52, 44, 36, 28,
	20, 12,  4, 27, 19, 11,  3
};

static const cc_u32 SDRM_des_skb[8][64]={
	{
		/* for C bits (numbered as per FIPS 46) 1 2 3 4 5 6 */
		0x00000000L,0x00000010L,0x20000000L,0x20000010L,
		0x00010000L,0x00010010L,0x20010000L,0x20010010L,
		0x00000800L,0x00000810L,0x20000800L,0x20000810L,
		0x00010800L,0x00010810L,0x20010800L,0x20010810L,
		0x00000020L,0x00000030L,0x20000020L,0x20000030L,
		0x00010020L,0x00010030L,0x20010020L,0x20010030L,
		0x00000820L,0x00000830L,0x20000820L,0x20000830L,
		0x00010820L,0x00010830L,0x20010820L,0x20010830L,
		0x00080000L,0x00080010L,0x20080000L,0x20080010L,
		0x00090000L,0x00090010L,0x20090000L,0x20090010L,
		0x00080800L,0x00080810L,0x20080800L,0x20080810L,
		0x00090800L,0x00090810L,0x20090800L,0x20090810L,
		0x00080020L,0x00080030L,0x20080020L,0x20080030L,
		0x00090020L,0x00090030L,0x20090020L,0x20090030L,
		0x00080820L,0x00080830L,0x20080820L,0x20080830L,
		0x00090820L,0x00090830L,0x20090820L,0x20090830L,
	},
	{
		/* for C bits (numbered as per FIPS 46) 7 8 10 11 12 13 */
		0x00000000L,0x02000000L,0x00002000L,0x02002000L,
		0x00200000L,0x02200000L,0x00202000L,0x02202000L,
		0x00000004L,0x02000004L,0x00002004L,0x02002004L,
		0x00200004L,0x02200004L,0x00202004L,0x02202004L,
		0x00000400L,0x02000400L,0x00002400L,0x02002400L,
		0x00200400L,0x02200400L,0x00202400L,0x02202400L,
		0x00000404L,0x02000404L,0x00002404L,0x02002404L,
		0x00200404L,0x02200404L,0x00202404L,0x02202404L,
		0x10000000L,0x12000000L,0x10002000L,0x12002000L,
		0x10200000L,0x12200000L,0x10202000L,0x12202000L,
		0x10000004L,0x12000004L,0x10002004L,0x12002004L,
		0x10200004L,0x12200004L,0x10202004L,0x12202004L,
		0x10000400L,0x12000400L,0x10002400L,0x12002400L,
		0x10200400L,0x12200400L,0x10202400L,0x12202400L,
		0x10000404L,0x12000404L,0x10002404L,0x12002404L,
		0x10200404L,0x12200404L,0x10202404L,0x12202404L,
	},
	{
		/* for C bits (numbered as per FIPS 46) 14 15 16 17 19 20 */
		0x00000000L,0x00000001L,0x00040000L,0x00040001L,
		0x01000000L,0x01000001L,0x01040000L,0x01040001L,
		0x00000002L,0x00000003L,0x00040002L,0x00040003L,
		0x01000002L,0x01000003L,0x01040002L,0x01040003L,
		0x00000200L,0x00000201L,0x00040200L,0x00040201L,
		0x01000200L,0x01000201L,0x01040200L,0x01040201L,
		0x00000202L,0x00000203L,0x00040202L,0x00040203L,
		0x01000202L,0x01000203L,0x01040202L,0x01040203L,
		0x08000000L,0x08000001L,0x08040000L,0x08040001L,
		0x09000000L,0x09000001L,0x09040000L,0x09040001L,
		0x08000002L,0x08000003L,0x08040002L,0x08040003L,
		0x09000002L,0x09000003L,0x09040002L,0x09040003L,
		0x08000200L,0x08000201L,0x08040200L,0x08040201L,
		0x09000200L,0x09000201L,0x09040200L,0x09040201L,
		0x08000202L,0x08000203L,0x08040202L,0x08040203L,
		0x09000202L,0x09000203L,0x09040202L,0x09040203L,
	},
	{
		/* for C bits (numbered as per FIPS 46) 21 23 24 26 27 28 */
		0x00000000L,0x00100000L,0x00000100L,0x00100100L,
		0x00000008L,0x00100008L,0x00000108L,0x00100108L,
		0x00001000L,0x00101000L,0x00001100L,0x00101100L,
		0x00001008L,0x00101008L,0x00001108L,0x00101108L,
		0x04000000L,0x04100000L,0x04000100L,0x04100100L,
		0x04000008L,0x04100008L,0x04000108L,0x04100108L,
		0x04001000L,0x04101000L,0x04001100L,0x04101100L,
		0x04001008L,0x04101008L,0x04001108L,0x04101108L,
		0x00020000L,0x00120000L,0x00020100L,0x00120100L,
		0x00020008L,0x00120008L,0x00020108L,0x00120108L,
		0x00021000L,0x00121000L,0x00021100L,0x00121100L,
		0x00021008L,0x00121008L,0x00021108L,0x00121108L,
		0x04020000L,0x04120000L,0x04020100L,0x04120100L,
		0x04020008L,0x04120008L,0x04020108L,0x04120108L,
		0x04021000L,0x04121000L,0x04021100L,0x04121100L,
		0x04021008L,0x04121008L,0x04021108L,0x04121108L,
	},
	{
		/* for D bits (numbered as per FIPS 46) 1 2 3 4 5 6 */
		0x00000000L,0x10000000L,0x00010000L,0x10010000L,
		0x00000004L,0x10000004L,0x00010004L,0x10010004L,
		0x20000000L,0x30000000L,0x20010000L,0x30010000L,
		0x20000004L,0x30000004L,0x20010004L,0x30010004L,
		0x00100000L,0x10100000L,0x00110000L,0x10110000L,
		0x00100004L,0x10100004L,0x00110004L,0x10110004L,
		0x20100000L,0x30100000L,0x20110000L,0x30110000L,
		0x20100004L,0x30100004L,0x20110004L,0x30110004L,
		0x00001000L,0x10001000L,0x00011000L,0x10011000L,
		0x00001004L,0x10001004L,0x00011004L,0x10011004L,
		0x20001000L,0x30001000L,0x20011000L,0x30011000L,
		0x20001004L,0x30001004L,0x20011004L,0x30011004L,
		0x00101000L,0x10101000L,0x00111000L,0x10111000L,
		0x00101004L,0x10101004L,0x00111004L,0x10111004L,
		0x20101000L,0x30101000L,0x20111000L,0x30111000L,
		0x20101004L,0x30101004L,0x20111004L,0x30111004L,
	},
	{
		/* for D bits (numbered as per FIPS 46) 8 9 11 12 13 14 */
		0x00000000L,0x08000000L,0x00000008L,0x08000008L,
		0x00000400L,0x08000400L,0x00000408L,0x08000408L,
		0x00020000L,0x08020000L,0x00020008L,0x08020008L,
		0x00020400L,0x08020400L,0x00020408L,0x08020408L,
		0x00000001L,0x08000001L,0x00000009L,0x08000009L,
		0x00000401L,0x08000401L,0x00000409L,0x08000409L,
		0x00020001L,0x08020001L,0x00020009L,0x08020009L,
		0x00020401L,0x08020401L,0x00020409L,0x08020409L,
		0x02000000L,0x0A000000L,0x02000008L,0x0A000008L,
		0x02000400L,0x0A000400L,0x02000408L,0x0A000408L,
		0x02020000L,0x0A020000L,0x02020008L,0x0A020008L,
		0x02020400L,0x0A020400L,0x02020408L,0x0A020408L,
		0x02000001L,0x0A000001L,0x02000009L,0x0A000009L,
		0x02000401L,0x0A000401L,0x02000409L,0x0A000409L,
		0x02020001L,0x0A020001L,0x02020009L,0x0A020009L,
		0x02020401L,0x0A020401L,0x02020409L,0x0A020409L,
	},
	{
		/* for D bits (numbered as per FIPS 46) 16 17 18 19 20 21 */
		0x00000000L,0x00000100L,0x00080000L,0x00080100L,
		0x01000000L,0x01000100L,0x01080000L,0x01080100L,
		0x00000010L,0x00000110L,0x00080010L,0x00080110L,
		0x01000010L,0x01000110L,0x01080010L,0x01080110L,
		0x00200000L,0x00200100L,0x00280000L,0x00280100L,
		0x01200000L,0x01200100L,0x01280000L,0x01280100L,
		0x00200010L,0x00200110L,0x00280010L,0x00280110L,
		0x01200010L,0x01200110L,0x01280010L,0x01280110L,
		0x00000200L,0x00000300L,0x00080200L,0x00080300L,
		0x01000200L,0x01000300L,0x01080200L,0x01080300L,
		0x00000210L,0x00000310L,0x00080210L,0x00080310L,
		0x01000210L,0x01000310L,0x01080210L,0x01080310L,
		0x00200200L,0x00200300L,0x00280200L,0x00280300L,
		0x01200200L,0x01200300L,0x01280200L,0x01280300L,
		0x00200210L,0x00200310L,0x00280210L,0x00280310L,
		0x01200210L,0x01200310L,0x01280210L,0x01280310L,
	},
	{
		/* for D bits (numbered as per FIPS 46) 22 23 24 25 27 28 */
		0x00000000L,0x04000000L,0x00040000L,0x04040000L,
		0x00000002L,0x04000002L,0x00040002L,0x04040002L,
		0x00002000L,0x04002000L,0x00042000L,0x04042000L,
		0x00002002L,0x04002002L,0x00042002L,0x04042002L,
		0x00000020L,0x04000020L,0x00040020L,0x04040020L,
		0x00000022L,0x04000022L,0x00040022L,0x04040022L,
		0x00002020L,0x04002020L,0x00042020L,0x04042020L,
		0x00002022L,0x04002022L,0x00042022L,0x04042022L,
		0x00000800L,0x04000800L,0x00040800L,0x04040800L,
		0x00000802L,0x04000802L,0x00040802L,0x04040802L,
		0x00002800L,0x04002800L,0x00042800L,0x04042800L,
		0x00002802L,0x04002802L,0x00042802L,0x04042802L,
		0x00000820L,0x04000820L,0x00040820L,0x04040820L,
		0x00000822L,0x04000822L,0x00040822L,0x04040822L,
		0x00002820L,0x04002820L,0x00042820L,0x04042820L,
		0x00002822L,0x04002822L,0x00042822L,0x04042822L,
	}
};

static const cc_u32 SDRM_DES_SPtrans[8][64]={
	//  nibble 0
	{
		0x02080800, 0x00080000, 0x02000002, 0x02080802, 0x02000000, 0x00080802, 0x00080002, 0x02000002,
		0x00080802, 0x02080800, 0x02080000, 0x00000802, 0x02000802, 0x02000000, 0x00000000, 0x00080002,
		0x00080000, 0x00000002, 0x02000800, 0x00080800, 0x02080802, 0x02080000, 0x00000802, 0x02000800,
		0x00000002, 0x00000800, 0x00080800, 0x02080002, 0x00000800, 0x02000802, 0x02080002, 0x00000000,
		0x00000000, 0x02080802, 0x02000800, 0x00080002, 0x02080800, 0x00080000, 0x00000802, 0x02000800,
		0x02080002, 0x00000800, 0x00080800, 0x02000002, 0x00080802, 0x00000002, 0x02000002, 0x02080000,
		0x02080802, 0x00080800, 0x02080000, 0x02000802, 0x02000000, 0x00000802, 0x00080002, 0x00000000,
		0x00080000, 0x02000000, 0x02000802, 0x02080800, 0x00000002, 0x02080002, 0x00000800, 0x00080802
	},
	//  nibble 1
	{
		0x40108010, 0x00000000, 0x00108000, 0x40100000, 0x40000010, 0x00008010, 0x40008000, 0x00108000,
		0x00008000, 0x40100010, 0x00000010, 0x40008000, 0x00100010, 0x40108000, 0x40100000, 0x00000010,
		0x00100000, 0x40008010, 0x40100010, 0x00008000, 0x00108010, 0x40000000, 0x00000000, 0x00100010,
		0x40008010, 0x00108010, 0x40108000, 0x40000010, 0x40000000, 0x00100000, 0x00008010, 0x40108010,
		0x00100010, 0x40108000, 0x40008000, 0x00108010, 0x40108010, 0x00100010, 0x40000010, 0x00000000,
		0x40000000, 0x00008010, 0x00100000, 0x40100010, 0x00008000, 0x40000000, 0x00108010, 0x40008010,
		0x40108000, 0x00008000, 0x00000000, 0x40000010, 0x00000010, 0x40108010, 0x00108000, 0x40100000,
		0x40100010, 0x00100000, 0x00008010, 0x40008000, 0x40008010, 0x00000010, 0x40100000, 0x00108000
	},
	//  nibble 2
	{
		0x04000001, 0x04040100, 0x00000100, 0x04000101, 0x00040001, 0x04000000, 0x04000101, 0x00040100,
		0x04000100, 0x00040000, 0x04040000, 0x00000001, 0x04040101, 0x00000101, 0x00000001, 0x04040001,
		0x00000000, 0x00040001, 0x04040100, 0x00000100, 0x00000101, 0x04040101, 0x00040000, 0x04000001,
		0x04040001, 0x04000100, 0x00040101, 0x04040000, 0x00040100, 0x00000000, 0x04000000, 0x00040101,
		0x04040100, 0x00000100, 0x00000001, 0x00040000, 0x00000101, 0x00040001, 0x04040000, 0x04000101,
		0x00000000, 0x04040100, 0x00040100, 0x04040001, 0x00040001, 0x04000000, 0x04040101, 0x00000001,
		0x00040101, 0x04000001, 0x04000000, 0x04040101, 0x00040000, 0x04000100, 0x04000101, 0x00040100,
		0x04000100, 0x00000000, 0x04040001, 0x00000101, 0x04000001, 0x00040101, 0x00000100, 0x04040000
	},
	//  nibble 3
	{
		0x00401008, 0x10001000, 0x00000008, 0x10401008, 0x00000000, 0x10400000, 0x10001008, 0x00400008,
		0x10401000, 0x10000008, 0x10000000, 0x00001008, 0x10000008, 0x00401008, 0x00400000, 0x10000000,
		0x10400008, 0x00401000, 0x00001000, 0x00000008, 0x00401000, 0x10001008, 0x10400000, 0x00001000,
		0x00001008, 0x00000000, 0x00400008, 0x10401000, 0x10001000, 0x10400008, 0x10401008, 0x00400000,
		0x10400008, 0x00001008, 0x00400000, 0x10000008, 0x00401000, 0x10001000, 0x00000008, 0x10400000,
		0x10001008, 0x00000000, 0x00001000, 0x00400008, 0x00000000, 0x10400008, 0x10401000, 0x00001000,
		0x10000000, 0x10401008, 0x00401008, 0x00400000, 0x10401008, 0x00000008, 0x10001000, 0x00401008,
		0x00400008, 0x00401000, 0x10400000, 0x10001008, 0x00001008, 0x10000000, 0x10000008, 0x10401000
	},
	//  nibble 4
	{
		0x08000000, 0x00010000, 0x00000400, 0x08010420, 0x08010020, 0x08000400, 0x00010420, 0x08010000,
		0x00010000, 0x00000020, 0x08000020, 0x00010400, 0x08000420, 0x08010020, 0x08010400, 0x00000000,
		0x00010400, 0x08000000, 0x00010020, 0x00000420, 0x08000400, 0x00010420, 0x00000000, 0x08000020,
		0x00000020, 0x08000420, 0x08010420, 0x00010020, 0x08010000, 0x00000400, 0x00000420, 0x08010400,
		0x08010400, 0x08000420, 0x00010020, 0x08010000, 0x00010000, 0x00000020, 0x08000020, 0x08000400,
		0x08000000, 0x00010400, 0x08010420, 0x00000000, 0x00010420, 0x08000000, 0x00000400, 0x00010020,
		0x08000420, 0x00000400, 0x00000000, 0x08010420, 0x08010020, 0x08010400, 0x00000420, 0x00010000,
		0x00010400, 0x08010020, 0x08000400, 0x00000420, 0x00000020, 0x00010420, 0x08010000, 0x08000020
	},
	//  nibble 5
	{
		0x80000040, 0x00200040, 0x00000000, 0x80202000, 0x00200040, 0x00002000, 0x80002040, 0x00200000,
		0x00002040, 0x80202040, 0x00202000, 0x80000000, 0x80002000, 0x80000040, 0x80200000, 0x00202040,
		0x00200000, 0x80002040, 0x80200040, 0x00000000, 0x00002000, 0x00000040, 0x80202000, 0x80200040,
		0x80202040, 0x80200000, 0x80000000, 0x00002040, 0x00000040, 0x00202000, 0x00202040, 0x80002000,
		0x00002040, 0x80000000, 0x80002000, 0x00202040, 0x80202000, 0x00200040, 0x00000000, 0x80002000,
		0x80000000, 0x00002000, 0x80200040, 0x00200000, 0x00200040, 0x80202040, 0x00202000, 0x00000040,
		0x80202040, 0x00202000, 0x00200000, 0x80002040, 0x80000040, 0x80200000, 0x00202040, 0x00000000,
		0x00002000, 0x80000040, 0x80002040, 0x80202000, 0x80200000, 0x00002040, 0x00000040, 0x80200040
	},

	//  nibble 6
	{
		0x00004000, 0x00000200, 0x01000200, 0x01000004, 0x01004204, 0x00004004, 0x00004200, 0x00000000,
		0x01000000, 0x01000204, 0x00000204, 0x01004000, 0x00000004, 0x01004200, 0x01004000, 0x00000204,
		0x01000204, 0x00004000, 0x00004004, 0x01004204, 0x00000000, 0x01000200, 0x01000004, 0x00004200,
		0x01004004, 0x00004204, 0x01004200, 0x00000004, 0x00004204, 0x01004004, 0x00000200, 0x01000000,
		0x00004204, 0x01004000, 0x01004004, 0x00000204, 0x00004000, 0x00000200, 0x01000000, 0x01004004,
		0x01000204, 0x00004204, 0x00004200, 0x00000000, 0x00000200, 0x01000004, 0x00000004, 0x01000200,
		0x00000000, 0x01000204, 0x01000200, 0x00004200, 0x00000204, 0x00004000, 0x01004204, 0x01000000,
		0x01004200, 0x00000004, 0x00004004, 0x01004204, 0x01000004, 0x01004200, 0x01004000, 0x00004004
	},
	//  nibble 7
	{
		0x20800080, 0x20820000, 0x00020080, 0x00000000, 0x20020000, 0x00800080, 0x20800000, 0x20820080,
		0x00000080, 0x20000000, 0x00820000, 0x00020080, 0x00820080, 0x20020080, 0x20000080, 0x20800000,
		0x00020000, 0x00820080, 0x00800080, 0x20020000, 0x20820080, 0x20000080, 0x00000000, 0x00820000,
		0x20000000, 0x00800000, 0x20020080, 0x20800080, 0x00800000, 0x00020000, 0x20820000, 0x00000080,
		0x00800000, 0x00020000, 0x20000080, 0x20820080, 0x00020080, 0x20000000, 0x00000000, 0x00820000,
		0x20800080, 0x20020080, 0x20020000, 0x00800080, 0x20820000, 0x00000080, 0x00800080, 0x20020000,
		0x20820080, 0x00800000, 0x20800000, 0x20000080, 0x00820000, 0x00020080, 0x20020080, 0x20800000,
		0x00000080, 0x20820000, 0x00820080, 0x00000000, 0x20000000, 0x20800080, 0x00020000, 0x00820080
	}
};

////////////////////////////////////////////////////////////////////////////
// Function Prototypes
////////////////////////////////////////////////////////////////////////////

#ifdef __cplusplus
extern "C" {
#endif

/*
 * @fn		int SDRM_DES_KeySched(cc_u8 *RoundKey, cc_u8 *UserKey, cc_u32 RKPos, cc_u32 RKStep)
 * @brief	Expand the cipher key into the encryption key schedule
 *
 * @param	RoundKey			[out]generated round key
 * @param	UserKey				[in]user key, 8 byte
 * @param	RKPos				[in]index of round key starts
 * @param	RKStep				[in]step for index
 *
 * @return	the number of rounds for the given cipher key size
 */
CRYPTOCORE_INTERNAL int SDRM_DES_KeySched(cc_u8 *RoundKey, cc_u8 *UserKey, cc_u32 RKPos, cc_u32 RKStep);

/*
 * @fn		int SDRM_DES_Encryption(cc_u32 RoundKey[][2], cc_u8 *msg, cc_u8 *out)
 * @brief	DES processing for one block
 *
 * @param	RoundKey			[in]expanded round key
 * @param	msg					[in]8 byte plaintext
 * @param	out					[out]8 byte ciphertext
 *
 * @return	CRYPTO_SUCCESS		if no error is occured
 */
CRYPTOCORE_INTERNAL int SDRM_DES_Encryption(cc_u32 RoundKey[][2], cc_u8 *msg, cc_u8 *out);

/*
 * @fn		int SDRM_DES64_Encryption(cc_u8 *cipherText, cc_u8 *plainText, cc_u8 *UserKey)
 * @brief	one block DES Encryption
 *
 * @param	cipherText	[out]encrypted text
 * @param	plainText	[in]plain text
 * @param	UserKey		[in]user key
 *
 * @return	CRYPTO_SUCCESS if success
 */
CRYPTOCORE_INTERNAL int SDRM_DES64_Encryption(cc_u8 *cipherText, cc_u8 *plainText, cc_u8 *UserKey);

/*
 * @fn		int SDRM_DES64_Decryption(cc_u8 *plainText, cc_u8 *cipherText, cc_u8 *UserKey)
 * @brief	one block DES Decryption
 *
 * @param	plainText	[out]decrypted text
 * @param	cipherText	[in]cipher text
 * @param	UserKey		[in]user key
 *
 * @return	CRYPTO_SUCCESS if success
 */
CRYPTOCORE_INTERNAL int SDRM_DES64_Decryption(cc_u8 *plainText, cc_u8 *cipherText, cc_u8 *UserKey);

#ifdef __cplusplus
}
#endif

#endif

/***************************** End of File *****************************/
